{=  Last modified                                                       =}
{=  $Date: 09.05.98 - 13:23:05 $                                      =}
unit Socks;
{$I DEFS.PAS}

interface
uses Windows,Winsock, Classes, SysUtils, messg;

type

  ESocketErrorException = class(Exception);
  ESocketTimeoutException = class(Exception);
  PLongint = ^Longint;

  TUDPSocket= class(TObject)
  private
    sock:TSocket;
  public
    Constructor Create(port:string);
    function SendBroadcast(var buf;len:integer;port:string):integer;
    function SendWinNuke(var buf;len:integer;adr,port:string):integer;
    function SendToAddr(var buf;len:integer;adr,port:string):integer;
    function ReceiveFrom(var buf;len:integer;var rip,rport,rname:string):integer;
  end;


  TClientSocket = class(TObject)
  private
    portn: Word;
    peerip: string;
    peername: string;
  public
    sock: tSocket;
    peer		:string;
    totell	:tstringlist;
    lastaccess :dword;
    property _socket: tSocket read sock;
    constructor Create;
    destructor Destroy;override;
    destructor DestroyEx; 
    procedure Connect(adr: string; PORT: string);
    procedure WaitForData;
    procedure WaitForDataTime(A: Word);
    procedure Disconnect;
    function read (var Buf; len: Word): Word;
    function write (var Buf; len: Word): Word;
  end;

  TServerSocket = class(TObject)
  protected
    sock: tSocket;
  private
    portn: Word;
    peerip: string;
    peername: string;
  public
    property FPeerIP: string read peerip;
    property FPeerName: string read peername;
    constructor Create(PORT: string);
    destructor Destroy;override;
    procedure Listen;
    procedure WaitForConnection;
    function AcceptConnection: TClientSocket;
    procedure Disconnect;
  end;

var
	WSAReady	:Boolean = False;
	Sox		:TList;
	MainPorts	:TList;
	MainPorts_:TStringList;
function GetPortByname(port,proto:string):integer;
implementation
function GetPortByname(port,proto:string):integer;
var
 PSE:PServEnt;
begin
 PSE := getservbyname(PChar(PORT), PChar(proto));
 if PSE = nil then begin
   {$IFDEF UCONSOLE}
   Warn(Format('cannot resolve service name %s; usign it as port number',[PORT]));
   {$ENDIF}
   Result:=StrToInt(PORT);
   Exit;
 end
 else begin
   Result := ntohs(PSE^.s_port);
   {$IFDEF UCONSOLE}
   Deb(Format('resolved; port number %d',[ntohs(PSE^.s_port)]), 2);
   {$ENDIF}
 end;
end;

{$IFDEF UCONSOLE}
procedure SocketError(const Msg: string);
var S: string;
begin
  S := Format('%s; error code %d',[Msg, WSAGetLastError]);
  ERR(S);
  raise ESocketErrorException.Create(S);
end;
{$ELSE}
procedure SocketError;
begin
  raise ESocketErrorException.Create('');
end;
{$ENDIF}

var
  WSAData: TWSAData;

procedure StartWSA;
begin
   {$IFDEF UCONSOLE}
   Deb('Starting WSA... ', 1);
   {$ENDIF}
  if WSAStartUp(257, WSAData) <> 0 then SocketError{$IFDEF UCONSOLE}('in socks.StartWSA: WSAStartUp() failed'){$ENDIF};
  WSAReady := True;
   {$IFDEF UCONSOLE}
  Deb('WSA: ' + WSAData.szDescription + ' ready', 1);
  {$ENDIF}
end;

{----------------------------- Server Socket implementation ---------------}

constructor TServerSocket.Create(PORT: string);
var
  sn : TSockAddrIn;
  PSE: PServEnt;
begin
  //assert(WSAReady, 'WSA not initialized');
  inherited Create;
  sock := INVALID_SOCKET; peername := ''; peerip := '';
  {$IFDEF UCONSOLE}
  Deb(Format('Creating master socket on port %s',[PORT]), 1);
  {$ENDIF}
  sock := socket(PF_INET, SOCK_STREAM, getprotobyname('tcp').p_proto);
  if sock = INVALID_SOCKET then SocketError{$IFDEF UCONSOLE}('in TServerSocket.Create: socket() failed'){$ENDIF};
   {$IFDEF UCONSOLE}
  Deb(Format('master socket: %d',[sock]), 2);
  {$ENDIF}
  sn.sin_family := AF_INET;
  sn.sin_addr.s_addr := INADDR_ANY;
   {$IFDEF UCONSOLE}
  Deb(Format('resolving port name %s',[PORT]), 2);
  {$ENDIF}
  PSE := getservbyname(PChar(PORT), PChar('tcp'));
  if PSE = nil then begin
    {$IFDEF UCONSOLE}
    Warn(Format('cannot resolv; usign port number %s',[PORT]));
    {$ENDIF}
    sn.sin_port := htons(StrToInt(PORT));
    portn := StrToInt(PORT);
  end
  else begin
    sn.sin_port := PSE^.s_port;
    portn := ntohs(PSE^.s_port);
   {$IFDEF UCONSOLE}
    Deb(Format('resolved; port number %d',[portn]), 2);
    {$ENDIF}
  end;
   {$IFDEF UCONSOLE}
  Deb('Binding socket', 2);
  {$ENDIF}
  if bind(sock, sn, sizeof(sn)) <> 0 then SocketError{$IFDEF UCONSOLE}('in TServerSocket.Create: bind() failed'){$ENDIF};
end;

destructor TServerSocket.Destroy;
begin
  if sock <> INVALID_SOCKET then Disconnect;
  inherited Destroy;
end;


procedure TServerSocket.Listen;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
   Deb(Format('Putting socket on port %d in listen state',[portn]), 1);
   {$ENDIF}
  if Winsock.Listen(sock, 5) <> 0 then SocketError{$IFDEF UCONSOLE}('in TServerSocket.Listen: listen() failed'){$ENDIF};
end;

procedure TServerSocket.WaitForConnection;
var
  fdset: TFDset;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
   Deb('Waiting for connection...', 1);
   {$ENDIF}
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  if Select(0, @fdset, nil, nil, nil) = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TServerSocket.WaitForConnect: select() failed'){$ENDIF};
end;

function TServerSocket.AcceptConnection;
var
  TmpSock: tSocket;
  sad    : TSockAddr;
  I      : Integer;
  phe    : PHostEnt;
  TmpCs  : TClientSocket;
begin
  Result := nil;
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb('Accepting connection...', 1);
  {$ENDIF}
  I := sizeof(TSockAddr);
  TmpSock := Accept(sock, sad, I);
  if TmpSock = INVALID_SOCKET then SocketError{$IFDEF UCONSOLE}('in TServerSocket.AcceptConnection: accept() failed'){$ENDIF};
  peerip := IntToStr(Ord(sad.sa_data[2])) + '.' + IntToStr(Ord(sad.sa_data[3])) + '.' +
    IntToStr(Ord(sad.sa_data[4])) + '.' + IntToStr(Ord(sad.sa_data[5]));

  {$ifdef DontAllowLocalsToConnect}
  If Sad.Sin_Addr.S_Addr = $0100007F then Begin
     closesocket(tmpsock);
	SocketError{$IFDEF UCONSOLE}('Local user messed up!! :-)'){$ENDIF};
  end;
  {$endif}

  {$IFDEF Resolve_Connector}
  phe := gethostbyaddr(@(sad.sa_data[2]), 4, PF_INET);
  if phe = nil then begin
    {$IFDEF UCONSOLE}
    Warn(Format('in TServerSocket.AcceptConnection: gethostbyaddr() failed; error code %d',[WSAGetLastError]));
    {$ENDIF}
    peername := '';
  end
  else peername := phe^.h_name;
  {$ELSE}
  peername := '';
  {$ENDIF}
  TmpCs := TClientSocket.Create;
  TmpCs.sock := TmpSock;
  TmpCs.peer := ' * ' + FPeerIP + ' (since ' + DateTimeToStr(Now) + ')';
  asm
	cli
  end;
  Sox.Add(TmpCs);
  asm
	sti
  end;
  Result := TmpCs;
end;

procedure TServerSocket.Disconnect;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Closing connection on socket %d...',[sock]), 1);
  {$ENDIF}
  if closesocket(sock) = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TClientSocket.disconnect: closesocket() failed'){$ENDIF};
  sock := INVALID_SOCKET;
  peername := '';
end;



{--------------------- End of Server Socket inplementation ---------------}



{--------------------- Client Socket implementation ----------------------}

constructor TClientSocket.Create;
begin
  //assert(WSAReady, 'WSA not initialized');
  inherited Create;
  lastaccess := gettickcount;
  totell := tstringlist.create;
  sock := INVALID_SOCKET; peername := ''; peerip := '';
end;


procedure TClientSocket.Connect(adr: string; PORT: string);
var phe: PHostEnt;
  sin   : TSockAddrIn;
  PSE   : PServEnt;
  szData: PChar;
begin
  //assert(WSAReady, 'WSA not initialized');
   {$IFDEF UCONSOLE}
  Deb('Creating client socket...', 1);
  {$ENDIF}
  sock := socket(PF_INET, SOCK_STREAM, getprotobyname('tcp').p_proto);
  if sock = INVALID_SOCKET then SocketError{$IFDEF UCONSOLE}('in TClientSocket socket() failed'){$ENDIF};
  sin.sin_family := AF_INET;
  PSE := getservbyname(PChar(PORT), PChar('tcp'));
  if PSE = nil then begin
    {$IFDEF UCONSOLE}
    Warn(Format('cannot resolve service name %s; usign is as port number',[PORT]));
    {$ENDIF}
    sin.sin_port := htons(StrToInt(PORT));
    portn := StrToInt(PORT);
  end
  else begin
    sin.sin_port := PSE^.s_port;
    portn := ntohs(PSE^.s_port);
   {$IFDEF UCONSOLE}
    Deb(Format('resolved; port number %d',[portn]), 2);
    {$ENDIF}
  end;
  szData := PChar(adr);
  sin.sin_addr.s_addr := inet_addr(szData);
  if sin.sin_addr.s_addr = INADDR_NONE then begin
   {$IFDEF UCONSOLE}
    Deb(Format('%s does not look like ip. Trying to use it as host name ',[adr]), 2);
    {$ENDIF}
    phe := gethostbyname(szData);
    if phe = nil then SocketError{$IFDEF UCONSOLE}(Format('Host name %s unknown',[adr])){$ENDIF};
   {$IFDEF UCONSOLE}
    Deb('resolved host name ', 2);
    {$ENDIF}
    sin.sin_addr.s_addr := Longint(PLongint(phe^.h_addr_list^)^);
  end;
   {$IFDEF UCONSOLE}
  Deb(Format('trying to connect to %s:%s',[adr, PORT]), 1);
  {$ENDIF}
  if Winsock.Connect(sock, sin, sizeof(sin)) <> 0 then SocketError{$IFDEF UCONSOLE}(Format('Cannot connect to %s:%s',[adr, PORT])){$ENDIF};
  peername := adr;
end;

procedure TClientSocket.WaitForData;
var
  fdset: TFDset;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Waiting for data on socket %d',[sock]), 1);
  {$ENDIF}
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  if Select(0, @fdset, nil, nil, nil) = SOCKET_ERROR then
    SocketError{$IFDEF UCONSOLE}('in TClientSocket.WaitForData: select() failed'){$ENDIF};
end;


procedure TClientSocket.WaitForDataTime;
var
  fdset: TFDset;  
  TM   : TTimeVal;
  Res  : Integer; 
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Waiting for data on socket %d %d microseconds',[sock, A]), 1);
  {$ENDIF}
  TM.tv_sec := A;
  TM.tv_usec := 0;
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  Res := Select(0, @fdset, nil, nil, @TM);
  if Res = 0 then raise ESocketTimeoutException.Create('');
  if Res = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TClientSocket.WaitForDataTime: select() failed'){$ENDIF};
end;


function TClientSocket.read;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Reading data on socket %d...',[sock]), 1);
  {$ENDIF}
  Result := recv(sock, Buf, len, 0);
  if result > 0 then
  lastaccess := gettickcount;
  if Result = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TClientSocket.read: recv() failed'){$ENDIF};
end;

function TClientSocket.write;
begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Writing data to socket %d...',[sock]), 1);
  {$ENDIF}
  lastaccess := gettickcount;
  Result := send(sock, Buf, len, 0);

  if Result = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TClientSocket.write: send() failed'){$ENDIF};
end;


procedure TClientSocket.Disconnect;
{type
  linger = record
    l_onoff: u_short;
    l_linger: u_short;
  end;
var
  iStatus: integer;
  lin: linger;
  linx: array[0..3] of char absolute lin;
begin
  shutdown(Sock,2);
  lin.l_onoff := 1;
  lin.l_linger := 0;
  setsockopt(Sock,SOL_SOCKET,SO_LINGER,linx,sizeof(lin));
  iStatus := closesocket(Sock);
  Sock := INVALID_SOCKET;
end;}


begin
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Closing connection on socket %d...',[sock]), 1);
  {$ENDIF}
  if closesocket(sock) = SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TClientSocket.disconnect: closesocket() failed'){$ENDIF};
  sock := INVALID_SOCKET;
  peername := '';
end;



destructor TClientSocket.Destroy;
var i:integer;
begin
  try
	i := sox.indexof(self);
	if i <> -1 then
	  sox.delete(i);
  except
  end;
  totell.free;
  if sock <> INVALID_SOCKET then Disconnect;
  inherited Destroy;
end;


destructor TClientSocket.DestroyEx;
var i:integer;
begin
  try
	i := sox.indexof(self);
	if i <> -1 then
	  sox.delete(i);
  except
  end;
  totell.free;
  inherited Destroy;
end;
{------------------- End of TClientSocket implementation--------------}



{------------------- Begin TUDPSocket implementation ----------------}
constructor TUDPSocket.Create;
var sin:TSockAddrIN;
begin
 //assert(WSAReady, 'WSA not initialized');
 inherited Create;
   {$IFDEF UCONSOLE}
 Deb('Creating UDP socket',1);
 {$ENDIF}
 sock:=INVALID_SOCKET;
 sock:=socket(PF_INET,SOCK_DGRAM,getprotobyname('udp').p_proto);
 if sock=INVALID_SOCKET then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.Create socket() failed'){$ENDIF};
 sin.sin_family := AF_INET;
 sin.sin_addr.s_addr := INADDR_ANY;
 sin.sin_port := htons(GetPortByName(PORT,'udp'));
 if bind(sock, sin, sizeof(sin))= SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.Create: bind() failed'){$ENDIF};
end;

function TUDPSocket.SendBroadcast(var buf;len:integer;port:string):integer;
var i,si:integer;
    sin:TSockAddrIn;
begin
  si:=SizeOf(integer);
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
   {$IFDEF UCONSOLE}
  Deb(Format('Sending broadcast on UDP socket %d...',[sock]), 1);
  {$ENDIF}
  if getsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast getsockopt() failed'){$ENDIF};
  if i=0 then begin
    i:=1;
    if setsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast setsockopt() failed'){$ENDIF};
  end;
 sin.sin_family:=AF_INET;
 sin.sin_addr.s_addr:=INADDR_BROADCAST;
 sin.sin_port:=htons(GetPortByName(port,'udp'));
 Result:=sendto(sock,buf,len,0,sin,SizeOf(sin));
 if Result=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadCast sendto() failed'){$ENDIF};
end;


function TUDPSocket.SendWinNuke(var buf;len:integer;adr,port:string):integer;
var i,si:integer;
    sin:TSockAddrIn;
    szData:Pchar;
    phe:PHostEnt;
begin
  si:=SizeOf(integer);
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  szData := PChar(adr);
  sin.sin_addr.s_addr := inet_addr(szData);
  if sin.sin_addr.s_addr = INADDR_NONE then begin
   {$IFDEF UCONSOLE}
    Deb(Format('%s does not look like ip. Trying to use it as host name ',[adr]), 2);
    {$ENDIF}
    phe := gethostbyname(szData);
    if phe = nil then SocketError{$IFDEF UCONSOLE}(Format('Host name %s unknown',[adr])){$ENDIF};
    {$IFDEF UCONSOLE}
    Deb('resolved host name ', 2);
    {$ENDIF}
    sin.sin_addr.s_addr := Longint(PLongint(phe^.h_addr_list^)^);
  end;
  {$IFDEF UCONSOLE}
  Deb(Format('Sending broadcast on UDP socket %s:%s on %d...',[adr,port,sock]), 1);
  {$ENDIF}
  if getsockopt(sock,SOL_SOCKET,SO_BROADCAST or SO_OOBINLINE,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast getsockopt() failed'){$ENDIF};
  if i=0 then begin
    i:=1;
    if setsockopt(sock,SOL_SOCKET,SO_BROADCAST or SO_OOBINLINE,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast setsockopt() failed'){$ENDIF};
  end;
 sin.sin_family:=AF_INET;
 sin.sin_addr.s_addr:=INADDR_BROADCAST;
 sin.sin_port:=htons(GetPortByName(port,'udp'));
 Result:=sendto(sock,buf,len,0,sin,SizeOf(sin));
 if Result=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadCast sendto() failed'){$ENDIF};
end;


function TUDPSocket.SendToAddr(var buf;len:integer;adr,port:string):integer;
var i,si:integer;
    sin:TSockAddrIn;
    szData:Pchar;
    phe:PHostEnt;
begin
  si:=SizeOf(integer);
  //assert(WSAReady, 'WSA not initialized');
  //assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  szData := PChar(adr);
  sin.sin_addr.s_addr := inet_addr(szData);
  if sin.sin_addr.s_addr = INADDR_NONE then begin
   {$IFDEF UCONSOLE}
    Deb(Format('%s does not look like ip. Trying to use it as host name ',[adr]), 2);
    {$ENDIF}
    phe := gethostbyname(szData);
    if phe = nil then SocketError{$IFDEF UCONSOLE}(Format('Host name %s unknown',[adr])){$ENDIF};
    {$IFDEF UCONSOLE}
    Deb('resolved host name ', 2);
    {$ENDIF}
    sin.sin_addr.s_addr := Longint(PLongint(phe^.h_addr_list^)^);
  end;
  {$IFDEF UCONSOLE}
  Deb(Format('Sending broadcast on UDP socket %s:%s on %d...',[adr,port,sock]), 1);
  {$ENDIF}
  if getsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast getsockopt() failed'){$ENDIF};
  if i=0 then begin
    i:=1;
    if setsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadcast setsockopt() failed'){$ENDIF};
  end;
 sin.sin_family:=AF_INET;
 sin.sin_addr.s_addr:=INADDR_BROADCAST;
 sin.sin_port:=htons(GetPortByName(port,'udp'));
 Result:=sendto(sock,buf,len,0,sin,SizeOf(sin));
 if Result=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.SendBroadCast sendto() failed'){$ENDIF};
end;

function TUDPSocket.ReceiveFrom(var buf;len:integer;var rip,rport,rname:string):integer;
var sin:TSockAddrIN;
    ssin:integer;
    phe:PHostEnt;
begin
  ssin:=SizeOf(TSockAddrIn);
  //assert(WSAReady, 'WSA not initialized');
  //assert(Self.sock <> INVALID_SOCKET, 'Socket operation on non socket');
  {$IFDEF UCONSOLE}
  Deb(Format('Receiving data on socket %d...',[sock]),1);
  {$ENDIF}
  Result:=winsock.recvfrom(sock,buf,len,0,sin,ssin);
  if Result=SOCKET_ERROR then SocketError{$IFDEF UCONSOLE}('in TUDPSocket.ReceiveFrom recvfrom() failed'){$ENDIF};
  rip := IntToStr(Ord(sin.sa_data[2])) + '.' + IntToStr(Ord(sin.sa_data[3])) + '.' +
    IntToStr(Ord(sin.sa_data[4])) + '.' + IntToStr(Ord(sin.sa_data[5]));
  phe:=gethostbyaddr(@sin.sa_data[2],4,PF_INET);
  {$IFDEF UCONSOLE}if phe=nil then Warn('in TUDPSocket.ReceiveFrom gethostbyaddr() failed');{$ENDIF}
  if phe<>nil then rname:=phe.h_name;
  rport:='';
  {$IFDEF UCONSOLE}
  Deb(Format('After proc sock is %d',[sock]),1);
  {$ENDIF}
end;

initialization
	StartWSA;
	Sox := TList.Create;
	MainPorts := TList.Create;
	MainPorts_ := TStringList.Create;
finalization
	If WSACleanup = SOCKET_ERROR then
		SocketError{$IFDEF UCONSOLE}('in finalization WSACleanup() failed'){$ENDIF};
End.
